@@ -104,7 +104,12 @@ static bool do_assignment(const word_T *word, const value_T *value)
     __attribute__((nonnull));
 static wchar_t *value_to_string(const value_T *value)
     __attribute__((nonnull));
-static long do_long_calculation(tokentype_T ttype, long v1, long v2);
+static bool do_binary_calculation(
+	evalinfo_T *info, tokentype_T ttype,
+	value_T *lhs, value_T *rhs, value_T *result)
+    __attribute__((nonnull));
+static long do_long_calculation1(tokentype_T ttype, long v1, long v2);
+static long do_long_calculation2(tokentype_T ttype, long v1, long v2);
 static double do_double_calculation(tokentype_T ttype, double v1, double v2);
 static long do_double_comparison(tokentype_T ttype, double v1, double v2);
 static void parse_conditional(evalinfo_T *info, value_T *result)
@@ -257,25 +262,9 @@ void parse_assignment(evalinfo_T *info, value_T *result)
 		parse_assignment(info, &rhs);
 		if (result->type == VT_VAR) {
 		    word_T saveword = result->v_var;
-		    switch (coerce_type(info, result, &rhs)) {
-			case VT_LONG:
-			    if (!fail_if_will_divide_by_zero(
-					ttype, &rhs, info, result))
-				result->v_long = do_long_calculation(ttype,
-					result->v_long, rhs.v_long);
-			    break;
-			case VT_DOUBLE:
-			    if (!fail_if_will_divide_by_zero(
-					ttype, &rhs, info, result))
-				result->v_double = do_double_calculation(ttype,
-					result->v_double, rhs.v_double);
-			    break;
-			case VT_INVALID:
-			    result->type = VT_INVALID;
-			    break;
-			case VT_VAR:
-			    assert(false);
-		    }
+		    if (!do_binary_calculation(
+				info, ttype, result, &rhs, result))
+			break;
 		    if (!do_assignment(&saveword, result))
 			info->error = true, result->type = VT_INVALID;
 		} else if (result->type != VT_INVALID) {
@@ -329,9 +318,65 @@ wchar_t *value_to_string(const value_T *value)
     assert(false);
 }
 
+/* Applies the binary operation defined by token `ttype' to operands `lhs' and
+ * `rhs'. The operands may be modified as a result of coercion. The result is
+ * assigned to `*result' unless there is an error.
+ * Returns true iff successful. */
+bool do_binary_calculation(
+	evalinfo_T *info, tokentype_T ttype,
+	value_T *lhs, value_T *rhs, value_T *result)
+{
+    switch (ttype) {
+	case TT_ASTER:  case TT_ASTEREQUAL:
+	case TT_SLASH:  case TT_SLASHEQUAL:
+	case TT_PERCENT:  case TT_PERCENTEQUAL:
+	case TT_PLUS:  case TT_PLUSEQUAL:
+	case TT_MINUS:  case TT_MINUSEQUAL:
+	    result->type = coerce_type(info, lhs, rhs);
+	    if (fail_if_will_divide_by_zero(ttype, rhs, info, result))
+		return false;
+	    switch (result->type) {
+		case VT_LONG:
+		    result->v_long = do_long_calculation1(
+			    ttype, lhs->v_long, rhs->v_long);
+		    break;
+		case VT_DOUBLE:
+		    result->v_double = do_double_calculation(
+			    ttype, lhs->v_double, rhs->v_double);
+		    break;
+		case VT_VAR:
+		    assert(false);
+		case VT_INVALID:
+		    break;
+	    }
+	    break;
+	case TT_LESSLESS:  case TT_LESSLESSEQUAL:
+	case TT_GREATERGREATER:  case TT_GREATERGREATEREQUAL:
+	case TT_AMP:  case TT_AMPEQUAL:
+	case TT_HAT:  case TT_HATEQUAL:
+	case TT_PIPE:  case TT_PIPEEQUAL:
+	    coerce_integer(info, lhs);
+	    coerce_integer(info, rhs);
+	    if (lhs->type == VT_LONG && rhs->type == VT_LONG) {
+		result->type = VT_LONG;
+		result->v_long =
+		    do_long_calculation2(ttype, lhs->v_long, rhs->v_long);
+	    } else {
+		result->type = VT_INVALID;
+	    }
+	    break;
+	case TT_EQUAL:
+	    *result = *rhs;
+	    break;
+	default:
+	    assert(false);
+    }
+    return true;
+}
+
 /* Does unary or binary long calculation according to the specified operator
  * token. Division by zero is not allowed. */
-long do_long_calculation(tokentype_T ttype, long v1, long v2)
+long do_long_calculation1(tokentype_T ttype, long v1, long v2)
 {
     switch (ttype) {
 	case TT_PLUS:  case TT_PLUSEQUAL:
@@ -344,6 +389,16 @@ long do_long_calculation(tokentype_T ttype, long v1, long v2)
 	    return v1 / v2;
 	case TT_PERCENT:  case TT_PERCENTEQUAL:
 	    return v1 % v2;
+	default:
+	    assert(false);
+    }
+}
+
+/* Does unary or binary long calculation according to the specified operator
+ * token. */
+long do_long_calculation2(tokentype_T ttype, long v1, long v2)
+{
+    switch (ttype) {
 	case TT_LESSLESS:  case TT_LESSLESSEQUAL:
 	    return v1 << v2;
 	case TT_GREATERGREATER:  case TT_GREATERGREATEREQUAL:
@@ -366,8 +421,6 @@ long do_long_calculation(tokentype_T ttype, long v1, long v2)
 	    return v1 ^ v2;
 	case TT_PIPE:  case TT_PIPEEQUAL:
 	    return v1 | v2;
-	case TT_EQUAL:
-	    return v2;
 	default:
 	    assert(false);
     }
@@ -546,14 +599,7 @@ void parse_inclusive_or(evalinfo_T *info, value_T *result)
 	    case TT_PIPE:
 		next_token(info);
 		parse_exclusive_or(info, &rhs);
-		coerce_integer(info, result);
-		coerce_integer(info, &rhs);
-		if (result->type == VT_LONG && rhs.type == VT_LONG) {
-		    result->v_long = do_long_calculation(ttype,
-			    result->v_long, rhs.v_long);
-		} else {
-		    result->type = VT_INVALID;
-		}
+		do_binary_calculation(info, TT_PIPE, result, &rhs, result);
 		break;
 	    default:
 		return;
@@ -573,14 +619,7 @@ void parse_exclusive_or(evalinfo_T *info, value_T *result)
 	    case TT_HAT:
 		next_token(info);
 		parse_and(info, &rhs);
-		coerce_integer(info, result);
-		coerce_integer(info, &rhs);
-		if (result->type == VT_LONG && rhs.type == VT_LONG) {
-		    result->v_long = do_long_calculation(ttype,
-			    result->v_long, rhs.v_long);
-		} else {
-		    result->type = VT_INVALID;
-		}
+		do_binary_calculation(info, TT_HAT, result, &rhs, result);
 		break;
 	    default:
 		return;
@@ -600,14 +639,7 @@ void parse_and(evalinfo_T *info, value_T *result)
 	    case TT_AMP:
 		next_token(info);
 		parse_equality(info, &rhs);
-		coerce_integer(info, result);
-		coerce_integer(info, &rhs);
-		if (result->type == VT_LONG && rhs.type == VT_LONG) {
-		    result->v_long = do_long_calculation(ttype,
-			    result->v_long, rhs.v_long);
-		} else {
-		    result->type = VT_INVALID;
-		}
+		do_binary_calculation(info, TT_AMP, result, &rhs, result);
 		break;
 	    default:
 		return;
@@ -632,7 +664,7 @@ void parse_equality(evalinfo_T *info, value_T *result)
 		parse_relational(info, &rhs);
 		switch (coerce_type(info, result, &rhs)) {
 		    case VT_LONG:
-			result->v_long = do_long_calculation(ttype,
+			result->v_long = do_long_calculation2(ttype,
 				result->v_long, rhs.v_long);
 			break;
 		    case VT_DOUBLE:
@@ -674,7 +706,7 @@ void parse_relational(evalinfo_T *info, value_T *result)
 		parse_shift(info, &rhs);
 		switch (coerce_type(info, result, &rhs)) {
 		    case VT_LONG:
-			result->v_long = do_long_calculation(ttype,
+			result->v_long = do_long_calculation2(ttype,
 				result->v_long, rhs.v_long);
 			break;
 		    case VT_DOUBLE:
@@ -709,14 +741,7 @@ void parse_shift(evalinfo_T *info, value_T *result)
 	    case TT_GREATERGREATER:
 		next_token(info);
 		parse_additive(info, &rhs);
-		coerce_integer(info, result);
-		coerce_integer(info, &rhs);
-		if (result->type == VT_LONG && rhs.type == VT_LONG) {
-		    result->v_long = do_long_calculation(ttype,
-			    result->v_long, rhs.v_long);
-		} else {
-		    result->type = VT_INVALID;
-		}
+		do_binary_calculation(info, ttype, result, &rhs, result);
 		break;
 	    default:
 		return;
@@ -739,21 +764,7 @@ void parse_additive(evalinfo_T *info, value_T *result)
 	    case TT_MINUS:
 		next_token(info);
 		parse_multiplicative(info, &rhs);
-		switch (coerce_type(info, result, &rhs)) {
-		    case VT_LONG:
-			result->v_long = do_long_calculation(ttype,
-				result->v_long, rhs.v_long);
-			break;
-		    case VT_DOUBLE:
-			result->v_double = do_double_calculation(ttype,
-				result->v_double, rhs.v_double);
-			break;
-		    case VT_INVALID:
-			result->type = VT_INVALID;
-			break;
-		    case VT_VAR:
-			assert(false);
-		}
+		do_binary_calculation(info, ttype, result, &rhs, result);
 		break;
 	    default:
 		return;
@@ -778,25 +789,7 @@ void parse_multiplicative(evalinfo_T *info, value_T *result)
 	    case TT_PERCENT:
 		next_token(info);
 		parse_prefix(info, &rhs);
-		switch (coerce_type(info, result, &rhs)) {
-		    case VT_LONG:
-			if (!fail_if_will_divide_by_zero(
-				    ttype, &rhs, info, result))
-			    result->v_long = do_long_calculation(ttype,
-				    result->v_long, rhs.v_long);
-			break;
-		    case VT_DOUBLE:
-			if (!fail_if_will_divide_by_zero(
-				    ttype, &rhs, info, result))
-			    result->v_double = do_double_calculation(ttype,
-				    result->v_double, rhs.v_double);
-			break;
-		    case VT_INVALID:
-			result->type = VT_INVALID;
-			break;
-		    case VT_VAR:
-			assert(false);
-		}
+		do_binary_calculation(info, ttype, result, &rhs, result);
 		break;
 	    default:
 		return;
