@@ -1,6 +1,6 @@
 /* Yash: yet another shell */
 /* expand.c: word expansion */
-/* (C) 2007-2015 magicant */
+/* (C) 2007-2017 magicant */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -43,26 +43,37 @@
 #include "yash.h"
 
 
+/* characters that have special meanings in brace expansion, quote removal, and
+ * globbing. When an unquoted expansion includes these characters, they are
+ * backslashed to protect from unexpected side effects in succeeding expansion
+ * steps. */
+#define CHARS_ESCAPED L"\\\"\'{,}"
+
 static bool expand_and_split_words(
 	const wordunit_T *restrict w, plist_T *restrict list)
     __attribute__((nonnull(2)));
 
 /* data passed between expansion functions */
-struct expand_word_T {
+struct expand_four_T {
     plist_T *valuelist, *splitlist;
     xwcsbuf_T valuebuf;
     xstrbuf_T splitbuf;
-    bool putempty;
+    bool zeroword;
 };
+/* When "$@" appears during expansion and there is no positional parameter, the
+ * `zeroword' flag is set so that the quoted empty word can be removed later. */
 
-static bool expand_word(
-	const wordunit_T *restrict w, tildetype_T tilde, bool quoted,
-	plist_T *restrict valuelist, plist_T *restrict splitlist)
-    __attribute__((nonnull(4)));
-static bool expand_word_inner(const wordunit_T *restrict w,
-	tildetype_T tilde, bool quoted, bool rec,
-	struct expand_word_T *restrict e)
+static bool expand_four_and_remove_quotes(
+	const wordunit_T *restrict w,
+	tildetype_T tilde, bool processquotes, bool escapeall,
+	plist_T *restrict valuelist)
     __attribute__((nonnull(5)));
+static bool expand_four(const wordunit_T *restrict w,
+	tildetype_T tilde, bool processquotes, bool escapeall, bool rec,
+	struct expand_four_T *restrict e)
+    __attribute__((nonnull(6)));
+static void fill_splitbuf(struct expand_four_T *e, bool splittable)
+    __attribute__((nonnull));
 
 static wchar_t *expand_tilde(const wchar_t **ss,
 	bool hasnextwordunit, tildetype_T tt)
@@ -71,10 +82,8 @@ static wchar_t *expand_tilde(const wchar_t **ss,
 enum indextype_T { IDX_NONE, IDX_ALL, IDX_CONCAT, IDX_NUMBER, };
 
 static bool expand_param(const paramexp_T *restrict p, bool indq,
-	struct expand_word_T *restrict e)
+	struct expand_four_T *restrict e)
     __attribute__((nonnull));
-static wchar_t *expand_param_simple(const paramexp_T *p)
-    __attribute__((nonnull,malloc,warn_unused_result));
 static enum indextype_T parse_indextype(const wchar_t *indexstr)
     __attribute__((nonnull,pure));
 static wchar_t *trim_wstring(wchar_t *s, ssize_t startindex, ssize_t endindex)
@@ -89,7 +98,9 @@ static void match_each(void **restrict slist, const wchar_t *restrict pattern,
 static void subst_each(void **restrict slist, const wchar_t *pattern,
 	const wchar_t *subst, paramexptype_T type)
     __attribute__((nonnull));
-static void **concatenate_values(void **values)
+static wchar_t *concatenate_values(void **values, bool escape)
+    __attribute__((nonnull,malloc,warn_unused_result));
+static void **concatenate_values_into_array(void **values, bool escape)
     __attribute__((nonnull,malloc,warn_unused_result));
 static void subst_length_each(void **slist)
     __attribute__((nonnull));
@@ -130,6 +141,10 @@ static inline void add_sq(
     __attribute__((nonnull));
 static wchar_t *escaped_wcspbrk(const wchar_t *s, const wchar_t *accept)
     __attribute__((nonnull));
+static wchar_t *escaped_remove(const wchar_t *s, const wchar_t *reject)
+    __attribute__((nonnull,malloc,warn_unused_result));
+static inline wchar_t *escaped_remove_free(wchar_t *s, const wchar_t *reject)
+    __attribute__((nonnull,malloc,warn_unused_result));
 
 static void glob_all(void **restrict patterns, plist_T *restrict list)
     __attribute__((nonnull));
@@ -210,12 +225,24 @@ bool expand_and_split_words(
     pl_init(&valuelist1);
     pl_init(&splitlist1);
 
+    struct expand_four_T expand;
+    expand.valuelist = &valuelist1;
+    expand.splitlist = &splitlist1;
+    wb_init(&expand.valuebuf);
+    sb_init(&expand.splitbuf);
+    expand.zeroword = false;
+
     /* four expansions (w -> list1) */
-    if (!expand_word(w, TT_SINGLE, false, &valuelist1, &splitlist1)) {
+    if (!expand_four(w, TT_SINGLE, true, false, false, &expand)) {
 	plfree(pl_toary(&valuelist1), free);
 	plfree(pl_toary(&splitlist1), free);
+	wb_destroy(&expand.valuebuf);
+	sb_destroy(&expand.splitbuf);
 	return false;
     }
+    assert(expand.valuebuf.length == expand.splitbuf.length);
+    pl_add(expand.valuelist, wb_towcs(&expand.valuebuf));
+    pl_add(expand.splitlist, sb_tostr(&expand.splitbuf));
 
     /* brace expansion (list1 -> list2) */
     if (shopt_braceexpand) {
@@ -231,39 +258,62 @@ bool expand_and_split_words(
     }
 
     /* field splitting (list2 -> list) */
+    size_t oldlength = list->length;
     fieldsplit_all(pl_toary(&valuelist2), pl_toary(&splitlist2), list);
+    assert(oldlength <= list->length);
+
+    /* empty field removal */
+    if (list->length - oldlength == 1) {
+	wchar_t *field = list->contents[oldlength];
+	if (field[0] == L'\0' ||
+		(expand.zeroword && wcscmp(field, L"\"\"") == 0)) {
+	    free(field);
+	    pl_remove(list, oldlength, 1);
+	}
+    }
+
+    /* quote removal */
+    for (size_t i = oldlength; i < list->length; i++)
+	list->contents[i] = escaped_remove_free(list->contents[i], L"\"\'");
+
     return true;
 }
 
 /* Expands a single word: the four expansions and quote removal.
  * This function doesn't perform brace expansion, field splitting, globbing and
  * unescaping.
+ * If `processquotes' is true, single- and double-quotations are recognized as
+ * quotes. Otherwise, they are treated like backslashed characters.
+ * If `escapeall' is true, the expanded words are all backslashed as if the
+ * entire expansion is quoted.
+ * If `processquotes' and `escapeall' are false, only backslashes not preceding
+ * any of $, `, \ are self-backslashed.
  * If successful, the resulting word is returned as a newly malloced string
  * that may include backslash escapes.
  * On error, an error message is printed and NULL is returned.
  * On error in a non-interactive shell, the shell exits. */
-wchar_t *expand_single(const wordunit_T *arg, tildetype_T tilde)
+wchar_t *expand_single(const wordunit_T *arg,
+	tildetype_T tilde, bool processquotes, bool escapeall)
 {
-    wchar_t *result;
     plist_T list;
     pl_init(&list);
 
-    if (!expand_word(arg, tilde, false, &list, NULL)) {
+    if (!expand_four_and_remove_quotes(
+		arg, tilde, processquotes, escapeall, &list)) {
 	maybe_exit_on_error();
 	plfree(pl_toary(&list), free);
 	return NULL;
     }
-    if (list.length != 1) {
-	/* concatenate multiple words to a single */
-	const wchar_t *ifs = getvar(L VAR_IFS);
-	wchar_t padding[] = { ifs != NULL ? ifs[0] : L' ', L'\0' };
-	result = joinwcsarray(list.contents, padding);
-	plfree(pl_toary(&list), free);
-    } else {
-	result = list.contents[0];
-	pl_destroy(&list);
-    }
-    return result;
+
+    return concatenate_values(pl_toary(&list), true);
+}
+
+/* Like `expand_single', but the result is unescaped (if successful). */
+wchar_t *expand_single_and_unescape(const wordunit_T *arg,
+	tildetype_T tilde, bool processquotes, bool escapeall)
+{
+    wchar_t *result = expand_single(arg, tilde, processquotes, escapeall);
+    return result == NULL ? NULL : unescapefree(result);
 }
 
 /* Expands a single word: the four expansions, glob, quote removal and unescape.
@@ -278,7 +328,7 @@ wchar_t *expand_single(const wordunit_T *arg, tildetype_T tilde)
  * On error in a non-interactive shell, the shell exits. */
 char *expand_single_with_glob(const wordunit_T *arg, tildetype_T tilde)
 {
-    wchar_t *exp = expand_single(arg, tilde);
+    wchar_t *exp = expand_single(arg, tilde, true, false);
     char *result;
 
     if (exp == NULL)
@@ -325,129 +375,79 @@ noglob:
     return result;
 }
 
-/* Performs expansions in a string: parameter/arithmetic/command expansions.
- * Brace expansion, field splitting and globbing are not performed.
- * If `esc' is true, backslashes preceding $, `, \ are removed. Otherwise,
- * no quotations are removed.
- * The result is a newly malloced string.
- * On error, an error message is printed and NULL is returned. */
-wchar_t *expand_string(const wordunit_T *w, bool esc)
-{
-    bool ok = true;
-    xwcsbuf_T buf;
-    wchar_t *s;
-
-    wb_init(&buf);
-    for (; w != NULL; w = w->next) {
-	switch (w->wu_type) {
-	case WT_STRING:
-	    for (const wchar_t *ss = w->wu_string; *ss != L'\0'; ss++) {
-		if (esc && ss[0] == L'\\' && ss[1] != L'\0'
-			&& wcschr(L"$`\\", ss[1]) != NULL) {
-		    ss++;
-		    if (*ss != L'\0')
-			wb_wccat(&buf, *ss);
-		} else {
-		    wb_wccat(&buf, *ss);
-		}
-	    }
-	    break;
-	case WT_PARAM:
-	    s = expand_param_simple(w->wu_param);
-	    goto cat_s;
-	case WT_CMDSUB:
-	    s = exec_command_substitution(&w->wu_cmdsub);
-	    goto cat_s;
-	case WT_ARITH:
-	    s = expand_single(w->wu_arith, TT_NONE);
-	    if (s != NULL)
-		s = evaluate_arithmetic(unescapefree(s));
-	cat_s:
-	    if (s != NULL) {
-		wb_catfree(&buf, s);
-	    } else {
-		ok = false;
-	    }
-	    break;
-	}
-    }
-    if (ok) {
-	return wb_towcs(&buf);
-    } else {
-	wb_destroy(&buf);
-	maybe_exit_on_error();
-	return NULL;
-    }
-}
-
 
 /********** Four Expansions **********/
 
 /* Performs the four expansions in the specified single word.
- * The four expansions are tilde expansion, parameter expansion, command
- * substitution and arithmetic expansion.
  * `w' is the word in which expansions occur.
  * `tilde' is type of tilde expansion that is performed.
- * If `quoted' is true, the expanded words are all backslashed as if the entire
- * expansion is quoted.
+ * If `processquotes' is true, single- and double-quotations are recognized as
+ * quotes. Otherwise, they are treated like backslashed characters.
+ * If `escapeall' is true, the expanded words are all backslashed as if the
+ * entire expansion is quoted.
+ * If `processquotes' and `escapeall' are false, only backslashes not preceding
+ * any of $, `, \ are self-backslashed.
  * The expanded word is added to `valuelist' as a newly malloced wide string.
- * The splittability string is added to `splitlist' if `splitlist' is non-NULL.
  * Single- or double-quoted characters are unquoted and backslashed.
- * In most cases, one string is added to each of `valuelist' and `splitlist'.
- * If the word contains "$@", the result may be any number of strings.
+ * In most cases, one string is added to `valuelist'. If the word contains "$@",
+ * however, any number of strings may be added.
  * The return value is true iff successful. */
-/* A splittability string is an array of Boolean values that specifies where
- * the word can be split in field splitting. The word can be split at the nth
- * character iff the nth value of the splittability string is non-zero. */
-bool expand_word(
-	const wordunit_T *restrict w, tildetype_T tilde, bool quoted,
-	plist_T *restrict valuelist, plist_T *restrict splitlist)
+bool expand_four_and_remove_quotes(
+	const wordunit_T *restrict w,
+	tildetype_T tilde, bool processquotes, bool escapeall,
+	plist_T *restrict valuelist)
 {
-    struct expand_word_T expand;
+    size_t oldlength = valuelist->length;
+    struct expand_four_T expand;
 
     expand.valuelist = valuelist;
     wb_init(&expand.valuebuf);
-    expand.splitlist = splitlist;
-    if (expand.splitlist != NULL)
-	sb_init(&expand.splitbuf);
-    expand.putempty = false;
+    expand.splitlist = NULL;
+    expand.zeroword = false;
 
-    bool ok = expand_word_inner(w, tilde, quoted, false, &expand);
+    bool ok = expand_four(w, tilde, processquotes, escapeall, false, &expand);
 
-    if (expand.splitlist != NULL)
-	assert(expand.valuebuf.length == expand.splitbuf.length);
-
-    /* A quoted empty word, if any, is added to the list here. It is indicated
-     * by the `putempty' flag that is set when a quote is found. */
-    if (expand.valuebuf.length > 0 || expand.putempty) {
-	pl_add(expand.valuelist, wb_towcs(&expand.valuebuf));
-	if (expand.splitlist != NULL)
-	    pl_add(expand.splitlist, sb_tostr(&expand.splitbuf));
-    } else {
+    /* remove empty word for "$@" if $# == 0 */
+    if (valuelist->length == oldlength && expand.zeroword &&
+	    wcscmp(expand.valuebuf.contents, L"\"\"") == 0)
 	wb_destroy(&expand.valuebuf);
-	if (expand.splitlist != NULL)
-	    sb_destroy(&expand.splitbuf);
-    }
+    else
+	pl_add(valuelist, wb_towcs(&expand.valuebuf));
+
+    /* quote removal */
+    for (size_t i = oldlength; i < valuelist->length; i++)
+	valuelist->contents[i] =
+	    escaped_remove_free(valuelist->contents[i], L"\"\'");
 
     return ok;
 }
 
 /* Performs the four expansions in the specified single word.
+ * The four expansions are tilde expansion, parameter expansion, command
+ * substitution, and arithmetic expansion.
  * `w' is the word in which expansions occur.
  * `tilde' specifies the type of tilde expansion that is performed.
- * If `quoted' is true, the expanded words are all backslashed as if the entire
- * expansion is quoted.
+ * If `processquotes' is true, single- and double-quotations are recognized as
+ * quotes. Otherwise, they are treated like backslashed characters.
+ * If `escapeall' is true, the expanded words are all backslashed as if the
+ * entire expansion is quoted.
+ * If `processquotes' and `escapeall' are false, only backslashes not preceding
+ * any of $, `, \ are self-backslashed.
  * `rec' must be true iff this expansion is part of another expansion.
  * `e->valuebuf' must be initialized before calling this function and is used to
  * expand the current word. If `w' expands to multiple words, the last word is
  * put in `e->valuebuf' and the others are inserted to `e->valuelist'.
  * The splittability strings are put in `e->splitbuf' and `e->splitlist'
  * accordingly if `e->splitlist' is non-NULL.
- * Single- or double-quoted characters are unquoted and backslashed.
+ * Single- and double-quotations remain in the resulting word. In addition,
+ * characters inside those quotations are backslashed.
  * The return value is true iff successful. */
-bool expand_word_inner(const wordunit_T *restrict w,
-	tildetype_T tilde, bool quoted, bool rec,
-	struct expand_word_T *restrict e)
+/* A splittability string is an array of Boolean values that specifies where
+ * the word can be split in field splitting. The word can be split at the nth
+ * character iff the nth value of the splittability string is non-zero. */
+bool expand_four(const wordunit_T *restrict w,
+	tildetype_T tilde, bool processquotes, bool escapeall, bool rec,
+	struct expand_four_T *restrict e)
 {
     bool ok = true;
     bool indq = false;  /* in a double quote? */
@@ -455,12 +455,7 @@ bool expand_word_inner(const wordunit_T *restrict w,
     const wchar_t *ss;
     wchar_t *s;
 
-#define FILL_SBUF(c)                                            \
-    do {                                                        \
-	if (e->splitlist != NULL)                               \
-	    sb_ccat_repeat(&e->splitbuf, c,                     \
-		    e->valuebuf.length - e->splitbuf.length);   \
-    } while (false)
+#define FILL_SBUF(s) fill_splitbuf(e, !indq && !escapeall && (s));
 #define FILL_SBUF_SPLITTABLE   FILL_SBUF(true)
 #define FILL_SBUF_UNSPLITTABLE FILL_SBUF(false)
 
@@ -478,19 +473,32 @@ bool expand_word_inner(const wordunit_T *restrict w,
 	    while (*ss != L'\0') {
 		switch (*ss) {
 		case L'"':
+		    if (!processquotes)
+			goto escape;
 		    indq = !indq;
-		    e->putempty |= indq;
+		    wb_wccat(&e->valuebuf, L'"');
+		    FILL_SBUF_UNSPLITTABLE;
 		    break;
 		case L'\'':
-		    if (indq)
-			goto default_case;
-		    e->putempty = true;
+		    if (!processquotes || indq)
+			goto escape;
+		    wb_wccat(&e->valuebuf, L'\'');
 		    add_sq(&ss, &e->valuebuf, true);
+		    wb_wccat(&e->valuebuf, L'\'');
 		    FILL_SBUF_UNSPLITTABLE;
 		    break;
 		case L'\\':
+		    if (!processquotes) {
+			if (!escapeall) {
+			    wchar_t c = ss[1];
+			    if (c == L'$' || c == L'`' || c == L'\\')
+				ss++;
+			}
+			goto escape;
+		    }
+
 		    if (indq && wcschr(CHARS_ESCAPABLE, ss[1]) == NULL) {
-			goto default_case;
+			goto escape;
 		    } else {
 			wb_wccat(&e->valuebuf, L'\\');
 			if (*++ss != L'\0')
@@ -511,32 +519,33 @@ bool expand_word_inner(const wordunit_T *restrict w,
 			continue;
 		    }
 		    /* falls thru! */
-		default:  default_case:
-		    if (indq || quoted)
+		default:
+		    if (indq || escapeall)
+escape:
 			wb_wccat(&e->valuebuf, L'\\');
 		    wb_wccat(&e->valuebuf, *ss);
-		    FILL_SBUF(rec && !indq && !quoted);
+		    FILL_SBUF(rec);
 		    break;
 		}
 		ss++;
 	    }
 	    break;
 	case WT_PARAM:
-	    if (!expand_param(w->wu_param, indq || quoted, e))
+	    if (!expand_param(w->wu_param, indq || escapeall, e))
 		ok = false;
 	    break;
 	case WT_CMDSUB:
 	    s = exec_command_substitution(&w->wu_cmdsub);
 	    goto cat_s;
 	case WT_ARITH:
-	    s = expand_single(w->wu_arith, TT_NONE);
+	    s = expand_single_and_unescape(w->wu_arith, TT_NONE, true, false);
 	    if (s != NULL)
-		s = evaluate_arithmetic(unescapefree(s));
+		s = evaluate_arithmetic(s);
 cat_s:
 	    if (s != NULL) {
 		wb_catfree(&e->valuebuf, escapefree(s,
-			    (indq || quoted) ? NULL : CHARS_ESCAPED));
-		FILL_SBUF(!(indq || quoted));
+			    (indq || escapeall) ? NULL : CHARS_ESCAPED));
+		FILL_SBUF_SPLITTABLE;
 	    } else {
 		ok = false;
 	    }
@@ -544,9 +553,23 @@ cat_s:
 	}
     }
 
+#undef FILL_SBUF_UNSPLITTABLE
+#undef FILL_SBUF_SPLITTABLE
+#undef FILL_SBUF
+
     return ok;
 }
 
+/* Appends to `e->splitbuf' as many `splittable' as needed to match the length
+ * with `e->valuebuf'. */
+void fill_splitbuf(struct expand_four_T *e, bool splittable)
+{
+    if (e->splitlist == NULL)
+	return;
+    sb_ccat_repeat(
+	    &e->splitbuf, splittable, e->valuebuf.length - e->splitbuf.length);
+}
+
 /* Performs tilde expansion.
  * `ss' is a pointer to a pointer to the tilde character. The pointer is
  * increased so that it points to the character right after the expanded string.
@@ -586,28 +609,23 @@ wchar_t *expand_tilde(const wchar_t **ss, bool hasnextwordunit, tildetype_T tt)
 	return NULL;
     }
     if (!posixly_correct) {
-	size_t INIT(index);
-	if (username[0] == L'+') {
-	    if (username[1] == L'\0') {
-		home = getvar(L VAR_PWD);
-		goto finish;
+	if (username[0] == L'+' && username[1] == L'\0') {
+	    home = getvar(L VAR_PWD);
+	    goto finish;
+	}
+	if (username[0] == L'-' && username[1] == L'\0') {
+	    home = getvar(L VAR_OLDPWD);
+	    goto finish;
+	}
 #if YASH_ENABLE_DIRSTACK
-	    } else if (parse_dirstack_index(username, &index, &home, false)
+	if (username[0] == L'+' || username[0] == L'-') {
+	    size_t index;
+	    if (parse_dirstack_index(username, &index, &home, false)
 		    && index != SIZE_MAX) {
 		goto finish;
-#endif
-	    }
-	} else if (username[0] == L'-') {
-	    if (username[1] == L'\0') {
-		home = getvar(L VAR_OLDPWD);
-		goto finish;
-#if YASH_ENABLE_DIRSTACK
-	    } else if (parse_dirstack_index(username, &index, &home, false)
-		    && index != SIZE_MAX) {
-		goto finish;
-#endif
 	    }
 	}
+#endif
     }
     home = get_home_directory(username, false);
 finish:
@@ -622,7 +640,7 @@ finish:
  * The result is put in `e'.
  * Returns true iff successful. */
 bool expand_param(const paramexp_T *restrict p, bool indq,
-	struct expand_word_T *restrict e)
+	struct expand_four_T *restrict e)
 {
     /* parse indices first */
     ssize_t startindex, endindex;
@@ -630,7 +648,8 @@ bool expand_param(const paramexp_T *restrict p, bool indq,
     if (p->pe_start == NULL) {
 	startindex = 0, endindex = SSIZE_MAX, indextype = IDX_NONE;
     } else {
-	wchar_t *start = expand_single(p->pe_start, TT_NONE);
+	wchar_t *start = expand_single_and_unescape(
+		p->pe_start, TT_NONE, true, false);
 	if (start == NULL)
 	    return false;
 	indextype = parse_indextype(start);
@@ -647,7 +666,8 @@ bool expand_param(const paramexp_T *restrict p, bool indq,
 	    if (p->pe_end == NULL) {
 		endindex = (startindex == -1) ? SSIZE_MAX : startindex;
 	    } else {
-		wchar_t *end = expand_single(p->pe_end, TT_NONE);
+		wchar_t *end = expand_single_and_unescape(
+			p->pe_end, TT_NONE, true, false);
 		if (end == NULL || !evaluate_index(end, &endindex))
 		    return false;
 	    }
@@ -667,7 +687,8 @@ bool expand_param(const paramexp_T *restrict p, bool indq,
     if (p->pe_type & PT_NEST) {
 	plist_T plist;
 	pl_init(&plist);
-	if (!expand_word(p->pe_nest, TT_NONE, true, &plist, NULL)) {
+	if (!expand_four_and_remove_quotes(
+		    p->pe_nest, TT_NONE, true, true, &plist)) {
 	    plfree(pl_toary(&plist), free);
 	    return false;
 	}
@@ -784,7 +805,7 @@ treat_array:
 	if (unset) {
 subst:
 	    plfree(values, free);
-	    return expand_word_inner(p->pe_subst, TT_SINGLE, indq, true, e);
+	    return expand_four(p->pe_subst, TT_SINGLE, true, indq, true, e);
 	}
 	break;
     case PT_ASSIGN:
@@ -806,10 +827,10 @@ subst:
 			p->pe_name);
 		return false;
 	    }
-	    subst = expand_single(p->pe_subst, TT_SINGLE);
+	    subst = expand_single_and_unescape(
+		    p->pe_subst, TT_SINGLE, true, false);
 	    if (subst == NULL)
 		return false;
-	    subst = unescapefree(subst);
 	    if (v.type != GV_ARRAY) {
 		assert(v.type == GV_NOTFOUND || v.type == GV_SCALAR);
 		if (!set_variable(
@@ -849,7 +870,7 @@ subst:
     wchar_t *match;
     switch (p->pe_type & PT_MASK) {
     case PT_MATCH:
-	match = expand_single(p->pe_match, TT_SINGLE);
+	match = expand_single(p->pe_match, TT_SINGLE, true, false);
 	if (match == NULL) {
 	    plfree(values, free);
 	    return false;
@@ -858,15 +879,14 @@ subst:
 	free(match);
 	break;
     case PT_SUBST:
-	match = expand_single(p->pe_match, TT_SINGLE);
-	subst = expand_single(p->pe_subst, TT_SINGLE);
+	match = expand_single(p->pe_match, TT_SINGLE, true, false);
+	subst = expand_single_and_unescape(p->pe_subst, TT_SINGLE, true, false);
 	if (match == NULL || subst == NULL) {
 	    free(match);
 	    free(subst);
 	    plfree(values, free);
 	    return false;
 	}
-	subst = unescapefree(subst);
 	subst_each(values, match, subst, p->pe_type);
 	free(match);
 	free(subst);
@@ -874,28 +894,24 @@ subst:
     }
 
     /* concatenate the elements of `values' */
-    if (concat)
-	values = concatenate_values(values);
+    if (concat && indq)
+	values = concatenate_values_into_array(values, false);
 
     /* PT_NUMBER */
     if (p->pe_type & PT_NUMBER)
 	subst_length_each(values);
 
-    /* concatenate the elements of `values' */
-    if (!indq)
-	values = concatenate_values(values);
-
     /* backslash escape */
     for (size_t i = 0; values[i] != NULL; i++)
 	values[i] = escapefree(values[i], indq ? NULL : CHARS_ESCAPED);
 
     /* add the elements of `values' to `e->valuelist' */
     if (values[0] == NULL) {
-	e->putempty = false;
+	e->zeroword = true;
     } else {
 	/* add the first element */
 	wb_catfree(&e->valuebuf, values[0]);
-	FILL_SBUF(!indq);
+	fill_splitbuf(e, !indq);
 	if (values[1] != NULL) {
 	    pl_add(e->valuelist, wb_towcs(&e->valuebuf));
 	    if (e->splitlist != NULL)
@@ -915,7 +931,7 @@ subst:
 	    wb_initwith(&e->valuebuf, values[i]);
 	    if (e->splitlist != NULL) {
 		sb_init(&e->splitbuf);
-		FILL_SBUF(!indq);
+		fill_splitbuf(e, !indq);
 	    }
 	}
     }
@@ -924,40 +940,6 @@ subst:
     return true;
 }
 
-/* Performs parameter expansion and returns the resulting word.
- * If multiple words result, they are concatenated into a single string, each
- * separated by a space.
- * If successful, returns a newly-malloced wide string.
- * On error, returns NULL. */
-wchar_t *expand_param_simple(const paramexp_T *p)
-{
-    plist_T valuelist;
-    struct expand_word_T expand;
-
-    expand.valuelist = pl_init(&valuelist);
-    wb_init(&expand.valuebuf);
-    expand.splitlist = NULL;
-    expand.putempty = false;
-
-    bool ok = expand_param(p, false, &expand);
-
-    /* A quoted empty word, if any, is added to the list here. It is indicated
-     * by the `putempty' flag, which is set when a quote is found. */
-    if (expand.valuebuf.length > 0 || expand.putempty) {
-	pl_add(expand.valuelist, wb_towcs(&expand.valuebuf));
-    } else {
-	wb_destroy(&expand.valuebuf);
-    }
-
-    for (size_t i = 0; i < valuelist.length; i++)
-	valuelist.contents[i] = unescapefree(valuelist.contents[i]);
-
-    void **results = pl_toary(expand.valuelist);
-    wchar_t *result = ok ? joinwcsarray(results, L" ") : NULL;
-    plfree(results, free);
-    return result;
-}
-
 /* Returns IDX_ALL, IDX_CONCAT, IDX_NUMBER if `indexstr' is L"@", L"*",
  * L"#" respectively. Otherwise returns IDX_NONE. */
 enum indextype_T parse_indextype(const wchar_t *indexstr)
@@ -1044,9 +1026,9 @@ void **trim_array(void **a, ssize_t startindex, ssize_t endindex)
 void print_subst_as_error(const paramexp_T *p)
 {
     if (p->pe_subst != NULL) {
-	wchar_t *subst = expand_single(p->pe_subst, TT_SINGLE);
+	wchar_t *subst = expand_single_and_unescape(
+		p->pe_subst, TT_SINGLE, true, false);
 	if (subst != NULL) {
-	    subst = unescapefree(subst);
 	    if (p->pe_type & PT_NEST)
 		xerror(0, "%ls", subst);
 	    else
@@ -1131,25 +1113,37 @@ void subst_each(void **restrict slist, const wchar_t *pattern,
 /* Concatenates the wide strings in the specified array.
  * Array `*values' must be a NULL-terminated array of pointers to wide strings.
  * The strings are concatenated into one, each separated by the first $IFS
- * character.
+ * character. The separators are backslashed if `escape' is true.
  * The array and its element strings are all freed in this function.
- * The return value is a pointer to a newly malloced NULL-terminated array that
- * contains exactly one element: a pointer to the newly malloced wide string
- * that is the result of concatenation. */
-void **concatenate_values(void **values)
+ * The return value is a pointer to the newly malloced wide string that is the
+ * result of concatenation. */
+wchar_t *concatenate_values(void **values, bool escape)
+{
+    wchar_t *first = values[0];
+    if (first != NULL && values[1] == NULL) {
+	// no actual concatenation needed
+	free(values);
+	return first;
+    }
+
+    const wchar_t *ifs = getvar(L VAR_IFS);
+    wchar_t padding[] = { L'\\', ifs != NULL ? ifs[0] : L' ', L'\0' };
+    wchar_t *result = joinwcsarray(values, escape ? padding : &padding[1]);
+    plfree(values, free);
+    return result;
+}
+
+/* Like `concatenate_values', but returns a pointer to a newly malloced
+ * NULL-terminated array containing the concatenated string. */
+void **concatenate_values_into_array(void **values, bool escape)
 {
-    /* If the array contains just one string, simply return it. */
     if (values[0] != NULL && values[1] == NULL)
 	return values;
 
-    const wchar_t *ifs = getvar(L VAR_IFS);
-    wchar_t padding[] = { ifs != NULL ? ifs[0] : L' ', L'\0' };
-    wchar_t *result = joinwcsarray(values, padding);
-    plfree(values, free);
-    values = xmallocn(2, sizeof *values);
-    values[0] = result;
-    values[1] = NULL;
-    return values;
+    void **results = xmallocn(2, sizeof *values);
+    results[0] = concatenate_values(values, escape);
+    results[1] = NULL;
+    return results;
 }
 
 /* Substitutes each string in the specified array with a string that contains
@@ -1298,7 +1292,7 @@ bool try_expand_brace_sequence(
 	return false;
     startlen = has_leading_zero(c, &sign) ? (dotp - c) : 0;
     errno = 0;
-    start = wcstol(c, &c, 0);
+    start = wcstol(c, &c, 10);
     if (errno != 0 || c != dotp)
 	return false;
 
@@ -1311,7 +1305,7 @@ bool try_expand_brace_sequence(
 	return false;
     endlen = has_leading_zero(c, &sign) ? (dotbracep - c) : 0;
     errno = 0;
-    end = wcstol(c, &c, 0);
+    end = wcstol(c, &c, 10);
     if (errno != 0 || c != dotbracep)
 	return false;
 
@@ -1323,7 +1317,7 @@ bool try_expand_brace_sequence(
 	if (bracep == NULL || c == bracep)
 	    return false;
 	errno = 0;
-	delta = wcstol(c, &c, 0);
+	delta = wcstol(c, &c, 10);
 	if (delta == 0 || errno != 0 || c != bracep)
 	    return false;
     } else {
@@ -1423,9 +1417,7 @@ void fieldsplit_all(void **restrict valuelist, void **restrict splitlist,
  * `s' is the word to split and freed in this function.
  * `split' is the splittability string corresponding to `s' and also freed.
  * The results are added to `dest' as newly-malloced wide strings.
- * `ifs' must not be NULL.
- * As a special case, if `s' is non-empty but only contains whitespaces from
- * `ifs', then no field is produced. */
+ * `ifs' must not be NULL. */
 void fieldsplit(wchar_t *restrict s, char *restrict split,
 	const wchar_t *restrict ifs, plist_T *restrict dest)
 {
@@ -1435,10 +1427,6 @@ void fieldsplit(wchar_t *restrict s, char *restrict split,
     extract_fields(s, split, true, ifs, &fields);
     assert(fields.length % 2 == 0);
 
-    if (s[0] != L'\0' && fields.length == 2 &&
-	    fields.contents[0] == fields.contents[1])
-	pl_clear(&fields, 0);
-
     for (size_t i = 0; i < fields.length; i += 2) {
 	const wchar_t *start = fields.contents[i], *end = fields.contents[i+1];
 	pl_add(dest, xwcsndup(start, end - start));
@@ -1467,9 +1455,13 @@ void fieldsplit(wchar_t *restrict s, char *restrict split,
  * corresponding character in the splittability string is non-zero. Refer to
  * POSIX for how whitespaces are treated in field splitting.
  *
- * If the last field is empty, the index is taken just after the last character
- * that is not an IFS-whitespace. */
-/* Split examples (assuming `ifs' = L" -")
+ * If an IFS non-whitespace delimits an empty field, the field is assumed just
+ * before the non-whitespace delimiter. The empty last field is removed if
+ * `shopt_emptylastfield' is false.
+ *
+ * The return value is a pointer to the end of the input string (but before
+ * trailing IFS whitespaces). */
+/* Split examples (assuming `ifs' = L" -" and `shopt_emptylastfield' is true)
  *   ""                  ->   ""
  *   "  "                ->   ""
  *   " abc 123 "         ->   "abc" "123"
@@ -1479,17 +1471,19 @@ void fieldsplit(wchar_t *restrict s, char *restrict split,
  *   "abc--123"          ->   "abc" "" "123"
  *   "abc - - 123"       ->   "abc" "" "123"
  */
-void extract_fields(const wchar_t *restrict s, const char *restrict split,
+wchar_t *extract_fields(const wchar_t *restrict s, const char *restrict split,
 	bool escaped, const wchar_t *restrict ifs, plist_T *restrict dest)
 {
     size_t index = 0;
+    size_t ifswhitestartindex;
+    size_t oldlen = dest->length;
 
     /* true when the currently skipping IFS whitespaces immediately follow a
      * previously split field. */
     bool afterfield = false;
 
     for (;;) {
-	size_t ifswhitestartindex = index;
+	ifswhitestartindex = index;
 	index += skip_ifs_whitespaces(&s[index], &split[index], escaped, ifs);
 
 	/* extract next field, if any */
@@ -1504,7 +1498,7 @@ void extract_fields(const wchar_t *restrict s, const char *restrict split,
 	/* Now the current char is either null or a IFS non-whitespace. */
 
 	if (!afterfield)
-	    add_empty_field(dest, &s[ifswhitestartindex]);
+	    add_empty_field(dest, &s[index]);
 
 	/* skip (only one) IFS non-whitespace */
 	size_t ifsstartindex = index;
@@ -1515,8 +1509,17 @@ void extract_fields(const wchar_t *restrict s, const char *restrict split,
 	}
 
 	/* Now the current char is null. We're done. */
-	return;
+	break;
     }
+
+    /* remove the empty last field */
+    size_t newlen = dest->length;
+    if (!shopt_emptylastfield && newlen - oldlen >= 2 * 2 &&
+	    dest->contents[newlen - 2] == dest->contents[newlen - 1])
+	pl_remove(dest, newlen - 2, 2);
+
+    assert(dest->length - oldlen >= 2);
+    return (wchar_t *) &s[ifswhitestartindex];
 }
 
 /* If `*s' is a (possibly escaped if `escaped') IFS character, returns the
@@ -1723,6 +1726,33 @@ wchar_t *escaped_wcspbrk(const wchar_t *s, const wchar_t *accept)
     return NULL;
 }
 
+/* Removes characters in `reject' from `s'.
+ * Backslash escapes in `s' are recognized. Escapes and escaped characters are
+ * kept in the result.
+ * The result is a newly malloced string. */
+wchar_t *escaped_remove(const wchar_t *s, const wchar_t *reject)
+{
+    xwcsbuf_T result;
+    wb_init(&result);
+    for (;;) {
+	const wchar_t *rejectchar = escaped_wcspbrk(s, reject);
+	if (rejectchar == NULL)
+	    break;
+	wb_ncat_force(&result, s, rejectchar - s);
+	s = rejectchar + 1;
+    }
+    wb_cat(&result, s);
+    return wb_towcs(&result);
+}
+
+/* Like `escaped_remove', but frees `s' before returning the result. */
+wchar_t *escaped_remove_free(wchar_t *s, const wchar_t *reject)
+{
+    wchar_t *result = escaped_remove(s, reject);
+    free(s);
+    return result;
+}
+
 
 /********** File Name Expansion (Glob) **********/
 
@@ -1802,7 +1832,7 @@ wchar_t *parse_and_expand_string(const wchar_t *s, const char *name, bool esc)
 
     if (!parse_string(&info, &word))
 	return NULL;
-    result = expand_string(word, esc);
+    result = expand_single_and_unescape(word, TT_NONE, false, !esc);
     wordfree(word);
     return result;
 }
