@@ -1,6 +1,6 @@
 /* Yash: yet another shell */
 /* job.c: job control */
-/* (C) 2007-2013 magicant */
+/* (C) 2007-2016 magicant */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -59,14 +59,16 @@ static size_t find_next_job(size_t numlimit);
 static void apply_curstop(void);
 static int calc_status(int status)
     __attribute__((const));
+static inline int calc_status_of_process(const process_T *p)
+    __attribute__((nonnull,pure));
 static wchar_t *get_job_name(const job_T *job)
     __attribute__((nonnull,warn_unused_result));
 static char *get_process_status_string(const process_T *p, bool *needfree)
     __attribute__((nonnull,malloc,warn_unused_result));
 static char *get_job_status_string(const job_T *job, bool *needfree)
     __attribute__((nonnull,malloc,warn_unused_result));
-static int print_job_status(
-	size_t jobnumber, _Bool changedonly, _Bool verbose, FILE *f)
+static int print_job_status(size_t jobnumber,
+	bool changedonly, bool verbose, bool remove_done, FILE *f)
     __attribute__((nonnull));
 static size_t get_jobnumber_from_name(const wchar_t *name)
     __attribute__((nonnull,pure));
@@ -78,6 +80,8 @@ static bool jobs_builtin_print_job(size_t jobnumber,
 	bool runningonly, bool stoppedonly);
 static int continue_job(size_t jobnumber, job_T *job, bool fg)
     __attribute__((nonnull));
+static int wait_for_job_by_jobspec(const wchar_t *jobspec)
+    __attribute__((nonnull));
 static bool wait_builtin_has_job(bool jobcontrol);
 
 
@@ -479,7 +483,7 @@ int wait_for_job(size_t jobnumber, bool return_on_stop,
  * The exit status is assigned to `laststatus' in any case. */
 wchar_t **wait_for_child(pid_t cpid, pid_t cpgid, bool return_on_stop)
 {
-    job_T *job = xmalloc(sizeof *job + sizeof *job->j_procs);
+    job_T *job = xmalloc(add(sizeof *job, sizeof *job->j_procs));
     job->j_pgid = cpgid;
     job->j_status = JS_RUNNING;
     job->j_statuschanged = false;
@@ -602,18 +606,30 @@ int calc_status(int status)
     assert(false);
 }
 
+/* Computes the exit status of the specified process.
+ * The process state must be JS_DONE or JS_STOPPED. */
+int calc_status_of_process(const process_T *p)
+{
+    int s = p->pr_statuscode;
+    return (p->pr_pid == 0) ? s : calc_status(s);
+}
+
 /* Computes the exit status of the specified job.
  * The job state must be JS_DONE or JS_STOPPED. */
 int calc_status_of_job(const job_T *job)
 {
     switch (job->j_status) {
     case JS_DONE:
-	if (job->j_procs[job->j_pcount - 1].pr_pid != 0)
-	    return calc_status(job->j_procs[job->j_pcount - 1].pr_statuscode);
-	else
-	    return job->j_procs[job->j_pcount - 1].pr_statuscode;
+	if (!shopt_pipefail)
+	    return calc_status_of_process(&job->j_procs[job->j_pcount - 1]);
+	for (size_t i = job->j_pcount; i-- > 0; ) {
+	    int status = calc_status_of_process(&job->j_procs[i]);
+	    if (status != Exit_SUCCESS)
+		return status;
+	}
+	return Exit_SUCCESS;
     case JS_STOPPED:
-	for (int i = job->j_pcount; --i >= 0; ) {
+	for (size_t i = job->j_pcount; i-- > 0; ) {
 	    if (job->j_procs[i].pr_status == JS_STOPPED)
 		return calc_status(job->j_procs[i].pr_statuscode);
 	}
@@ -711,22 +727,24 @@ char *get_job_status_string(const job_T *job, bool *needfree)
 }
 
 /* Prints the status of the specified job.
- * Finished jobs are removed from the job list after the status is printed.
+ * If `remove_done' is true, finished jobs are removed from the job list after
+ * the status is printed.
  * If the specified job doesn't exist, nothing is printed (it isn't an error).
  * If `changedonly' is true, the job is printed only if the `j_statuschanged'
  * flag is true.
  * If `verbose' is true, the status is printed in the process-wise format rather
  * than the usual job-wise format.
  * Returns zero if successful. Returns errno if `fprintf' failed. */
-int print_job_status(size_t jobnumber, bool changedonly, bool verbose, FILE *f)
+int print_job_status(size_t jobnumber,
+	bool changedonly, bool verbose, bool remove_done, FILE *f)
 {
     int result = 0;
 
     job_T *job = get_job(jobnumber);
-    if (job == NULL
-	    || (changedonly && !job->j_statuschanged)
-	    || job->j_nonotify)
+    if (job == NULL || job->j_nonotify)
 	return result;
+    if (changedonly && !job->j_statuschanged)
+	goto done;
 
     char current;
     if      (jobnumber == current_jobnumber)  current = '+';
@@ -782,7 +800,8 @@ int print_job_status(size_t jobnumber, bool changedonly, bool verbose, FILE *f)
 	}
     }
     job->j_statuschanged = false;
-    if (job->j_status == JS_DONE)
+done:
+    if (remove_done && job->j_status == JS_DONE)
 	remove_job(jobnumber);
 
     return result;
@@ -793,7 +812,7 @@ void print_job_status_all(void)
 {
     apply_curstop();
     for (size_t i = 1; i < joblist.length; i++)
-	print_job_status(i, true, false, stderr);
+	print_job_status(i, true, false, false, stderr);
 }
 
 /* If the shell is interactive and the specified job has been killed by a
@@ -1056,7 +1075,7 @@ bool jobs_builtin_print_job(size_t jobnumber,
 	int result = printf("%jd\n", (intmax_t) job->j_pgid);
 	err = (result >= 0) ? 0 : errno;
     } else {
-	err = print_job_status(jobnumber, changedonly, verbose, stdout);
+	err = print_job_status(jobnumber, changedonly, verbose, true, stdout);
     }
     if (err != 0) {
 	xerror(err, Ngt("cannot print to the standard output"));
@@ -1134,6 +1153,8 @@ int fg_builtin(int argc, void **argv)
 	if (current_jobnumber == 0 ||
 		(job = joblist.contents[current_jobnumber])->j_legacy) {
 	    xerror(0, Ngt("there is no current job"));
+	} else if (job->j_pgid == 0) {
+	    xerror(0, Ngt("the current job is not a job-controlled job"));
 	} else {
 	    status = continue_job(current_jobnumber, job, fg);
 	}
@@ -1155,9 +1176,9 @@ int continue_job(size_t jobnumber, job_T *job, bool fg)
 
     wchar_t *name = get_job_name(job);
     if (fg && posixly_correct)
-	printf("%ls\n", name);
+	xprintf("%ls\n", name);
     else
-	printf("[%zu] %ls\n", jobnumber, name);
+	xprintf("[%zu] %ls\n", jobnumber, name);
     if (name != job->j_procs[0].pr_name)
 	free(name);
 
@@ -1252,47 +1273,16 @@ int wait_builtin(int argc, void **argv)
 	}
     }
 
-    job_T *job;
     if (xoptind < argc) {
 	/* wait for the specified jobs */
-	do {
-	    const wchar_t *jobspec = ARGV(xoptind);
-	    size_t jobnumber;
-	    if (jobspec[0] == L'%') {
-		jobnumber = get_jobnumber_from_name(&jobspec[1]);
-	    } else {
-		long pid;
-		if (!xwcstol(jobspec, 10, &pid) || pid < 0) {
-		    xerror(0, Ngt("`%ls' is not a valid job specification"),
-			    jobspec);
-		    continue;
-		}
-		jobnumber = get_jobnumber_from_pid(pid);
+	for (; xoptind < argc; xoptind++) {
+	    int jobstatus = wait_for_job_by_jobspec(ARGV(xoptind));
+	    if (jobstatus < 0) {
+		status = -jobstatus;
+		break;
 	    }
-	    if (jobnumber >= joblist.length) {
-		xerror(0, Ngt("job specification `%ls' is ambiguous"),
-			ARGV(xoptind));
-	    } else if (jobnumber == 0
-		    || (job = joblist.contents[jobnumber]) == NULL
-		    || job->j_legacy) {
-		status = Exit_NOTFOUND;
-	    } else {
-		status = wait_for_job(jobnumber, jobcontrol, jobcontrol, true);
-		if (status == 0) {
-		    status = calc_status_of_job(job);
-		} else {
-		    assert(TERMSIGOFFSET >= 128);
-		    status += TERMSIGOFFSET;
-		    break;
-		}
-		if (job->j_status != JS_RUNNING) {
-		    if (jobcontrol && is_interactive_now && !posixly_correct)
-			print_job_status(jobnumber, false, false, stdout);
-		    else if (job->j_status == JS_DONE)
-			remove_job(jobnumber);
-		}
-	    }
-	} while (++xoptind < argc);
+	    status = jobstatus;
+	}
     } else {
 	/* wait for all jobs */
 	while (wait_builtin_has_job(jobcontrol)) {
@@ -1305,11 +1295,52 @@ int wait_builtin(int argc, void **argv)
 	}
     }
 
-    if (status != 0)
-	return status;
     if (yash_error_message_count != 0)
 	return Exit_FAILURE;
-    return Exit_SUCCESS;
+    return status;
+}
+
+/* Finds a job specified by the argument and waits for it.
+ * Returns a negated exit status if interrupted. */
+int wait_for_job_by_jobspec(const wchar_t *jobspec)
+{
+    size_t jobnumber;
+    if (jobspec[0] == L'%') {
+	jobnumber = get_jobnumber_from_name(&jobspec[1]);
+    } else {
+	long pid;
+	if (!xwcstol(jobspec, 10, &pid) || pid < 0) {
+	    xerror(0, Ngt("`%ls' is not a valid job specification"), jobspec);
+	    return Exit_FAILURE;
+	}
+	jobnumber = get_jobnumber_from_pid(pid);
+    }
+    if (jobnumber >= joblist.length) {
+	xerror(0, Ngt("job specification `%ls' is ambiguous"), jobspec);
+	return Exit_FAILURE;
+    }
+
+    job_T *job;
+    if (jobnumber == 0
+	    || (job = joblist.contents[jobnumber]) == NULL
+	    || job->j_legacy)
+	return Exit_NOTFOUND;
+
+    int signal = wait_for_job(jobnumber,
+	    doing_job_control_now, doing_job_control_now, true);
+    if (signal != 0) {
+	assert(TERMSIGOFFSET >= 128);
+	return -(signal + TERMSIGOFFSET);
+    }
+
+    int status = calc_status_of_job(job);
+    if (job->j_status != JS_RUNNING) {
+	if (doing_job_control_now && is_interactive_now && !posixly_correct)
+	    print_job_status(jobnumber, false, false, true, stdout);
+	else if (job->j_status == JS_DONE)
+	    remove_job(jobnumber);
+    }
+    return status;
 }
 
 /* Checks if the shell has any job to wait for. */
@@ -1318,7 +1349,7 @@ bool wait_builtin_has_job(bool jobcontrol)
     /* print/remove already-finished jobs */
     if (jobcontrol && is_interactive_now && !posixly_correct) {
 	for (size_t i = 1; i < joblist.length; i++)
-	    print_job_status(i, true, false, stdout);
+	    print_job_status(i, true, false, true, stdout);
     } else {
 	for (size_t i = 1; i < joblist.length; i++) {
 	    job_T *job = joblist.contents[i];
