@@ -1,6 +1,6 @@
 /* Yash: yet another shell */
 /* sig.c: signal handling */
-/* (C) 2007-2015 magicant */
+/* (C) 2007-2016 magicant */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -56,7 +56,7 @@
 /* About the shell's signal handling:
  *
  * Yash always catches SIGCHLD.
- * When job control is active, SIGTSTP is ignored.
+ * When job control is active, SIGTTIN, SIGTTOU, and SIGTSTP are ignored.
  * If the shell is interactive, SIGTERM and SIGQUIT are ignored and SIGINT and
  * SIGWINCH are caught.
  * Trapped signals are also caught.
@@ -293,8 +293,8 @@ static volatile sig_atomic_t sigwinch_received;
 
 /* true iff SIGCHLD is handled. */
 static bool main_handler_set = false;
-/* true iff SIGTSTP is ignored. */
-static bool job_handler_set = false;
+/* true iff SIGTTIN, SIGTTOU, and SIGTSTP are ignored. */
+static bool job_handlers_set = false;
 /* true iff SIGTERM, SIGINT, SIGQUIT and SIGWINCH are ignored/handled. */
 static bool interactive_handlers_set = false;
 
@@ -314,8 +314,10 @@ void set_signals(void)
 {
     sigset_t block = trapped_signals;
 
-    if (!job_handler_set && doing_job_control_now) {
-	job_handler_set = true;
+    if (!job_handlers_set && doing_job_control_now) {
+	job_handlers_set = true;
+	set_special_handler(SIGTTIN, SIG_IGN);
+	set_special_handler(SIGTTOU, SIG_IGN);
 	set_special_handler(SIGTSTP, SIG_IGN);
     }
 
@@ -347,8 +349,10 @@ void set_signals(void)
  * If `leave' is false, the setting for SIGCHLD are not restored. */
 void restore_signals(bool leave)
 {
-    if (job_handler_set) {
-	job_handler_set = false;
+    if (job_handlers_set) {
+	job_handlers_set = false;
+	reset_special_handler(SIGTTIN, SIG_IGN, leave);
+	reset_special_handler(SIGTTOU, SIG_IGN, leave);
 	reset_special_handler(SIGTSTP, SIG_IGN, leave);
     }
     if (interactive_handlers_set) {
@@ -372,15 +376,19 @@ void restore_signals(bool leave)
     }
 }
 
-/* Re-sets the signal handler for SIGTSTP according to the current
- * `doing_job_control_now' and `job_handler_set'. */
+/* Re-sets the signal handler for SIGTTIN, SIGTTOU, and SIGTSTP according to the
+ * current `doing_job_control_now' and `job_handlers_set'. */
 void reset_job_signals(void)
 {
-    if (doing_job_control_now && !job_handler_set) {
-	job_handler_set = true;
+    if (doing_job_control_now && !job_handlers_set) {
+	job_handlers_set = true;
+	set_special_handler(SIGTTIN, SIG_IGN);
+	set_special_handler(SIGTTOU, SIG_IGN);
 	set_special_handler(SIGTSTP, SIG_IGN);
-    } else if (!doing_job_control_now && job_handler_set) {
-	job_handler_set = false;
+    } else if (!doing_job_control_now && job_handlers_set) {
+	job_handlers_set = false;
+	reset_special_handler(SIGTTIN, SIG_IGN, false);
+	reset_special_handler(SIGTTOU, SIG_IGN, false);
 	reset_special_handler(SIGTSTP, SIG_IGN, false);
     }
 }
@@ -570,15 +578,15 @@ int wait_for_sigchld(bool interruptible, bool return_on_trap)
  * `handle_sigchld' and `handle_sigwinch' are called to handle SIGCHLD and
  * SIGWINCH that are caught while waiting.
  * If `trap' is true, traps are also handled while waiting and the
- * `sigint_received' flag is cleared when this function returns.
+ * `sigint_received' flag is cleared when this function returns. The return
+ * value can be W_INTERRUPTED only if `trap' is true. Any SIGINT received before
+ * entering this function is ignored.
  * The maximum time length of wait is specified by `timeout' in milliseconds.
  * If `timeout' is negative, the wait time is unlimited.
  * If the wait is interrupted by a signal, this function will re-wait for the
  * specified timeout, which means that this function may wait for a time length
- * longer than the specified timeout.
- * This function returns true if the input is ready or false if an error
- * occurred or it timed out. */
-bool wait_for_input(int fd, bool trap, int timeout)
+ * longer than the specified timeout. */
+enum wait_for_input_T wait_for_input(int fd, bool trap, int timeout)
 {
     sigset_t ss;
     struct timespec to;
@@ -587,9 +595,12 @@ bool wait_for_input(int fd, bool trap, int timeout)
     assert(fd >= 0);
     if (fd >= FD_SETSIZE) {
 	xerror(0, Ngt("too many files are opened for yash to handle"));
-	return false;
+	return W_ERROR;
     }
 
+    if (trap)
+	sigint_received = false;
+
     ss = accept_sigmask;
     sigdelset(&ss, SIGCHLD);
     if (interactive_handlers_set) {
@@ -614,19 +625,28 @@ bool wait_for_input(int fd, bool trap, int timeout)
 #if YASH_ENABLE_LINEEDIT && defined SIGWINCH
 	handle_sigwinch();
 #endif
+	if (trap && sigint_received) {
+	    sigint_received = false;
+	    return W_INTERRUPTED;
+	}
 
 	fd_set fdset;
 	FD_ZERO(&fdset);
 	FD_SET(fd, &fdset);
-	if (pselect(fd + 1, &fdset, NULL, NULL, top, &ss) >= 0) {
-	    if (trap)
-		sigint_received = false;
-	    return FD_ISSET(fd, &fdset);
-	} else {
-	    if (errno != EINTR) {
-		xerror(errno, "pselect");
-		return false;
-	    }
+
+	int count = pselect(fd + 1, &fdset, NULL, NULL, top, &ss);
+
+	if (trap && sigint_received) {
+	    sigint_received = false;
+	    return W_INTERRUPTED;
+	}
+
+	if (count >= 0)
+	    return FD_ISSET(fd, &fdset) ? W_READY : W_TIMED_OUT;
+
+	if (errno != EINTR) {
+	    xerror(errno, "pselect");
+	    return W_ERROR;
 	}
     }
 }
@@ -708,10 +728,12 @@ int handle_traps(void)
 		    le_suspend_readline();
 #endif
 		    struct execstate_T *execstate = save_execstate();
+		    reset_execstate(true);
 		    signum = handled_signal = s->no;
 		    command = xwcsdup(command);
 		    exec_wcs(command, "trap", false);
 		    free(command);
+		    cancel_return();
 		    restore_execstate(execstate);
 		    laststatus = savelaststatus;
 		}
@@ -730,10 +752,12 @@ int handle_traps(void)
 		    le_suspend_readline();
 #endif
 		    struct execstate_T *execstate = save_execstate();
+		    reset_execstate(true);
 		    signum = handled_signal = sigrtmin + i;
 		    command = xwcsdup(command);
 		    exec_wcs(command, "trap", false);
 		    free(command);
+		    cancel_return();
 		    restore_execstate(execstate);
 		    laststatus = savelaststatus;
 		}
@@ -770,7 +794,7 @@ void execute_exit_trap(void)
 #endif
 	savelaststatus = laststatus;
 	command = xwcsdup(command);
-	reset_execstate();
+	reset_execstate(true);
 	exec_wcs(command, "EXIT trap", false);
 	free(command);
 	savelaststatus = -1;
@@ -870,8 +894,10 @@ void set_trap(int signum, const wchar_t *command)
 	    if (interactive_handlers_set)
 		return;
 	    break;
+	case SIGTTIN:
+	case SIGTTOU:
 	case SIGTSTP:
-	    if (job_handler_set)
+	    if (job_handlers_set)
 		goto default_ignore;
 	    break;
 	case SIGTERM:
@@ -905,7 +931,8 @@ bool is_ignored(int signum)
     if (signum == 0)
 	return false;
 
-    if (doing_job_control_now && signum == SIGTSTP)
+    if (doing_job_control_now &&
+	    (signum == SIGTTIN || signum == SIGTTOU || signum == SIGTSTP))
 	return sigismember(&ignored_signals, signum);
 
     struct sigaction action;
@@ -1007,13 +1034,6 @@ void set_interrupted(void)
     sigint_received = true;
 }
 
-/* Resets the `sigint_received' flag. */
-/* This function should be called just after calling `handle_signals'. */
-void reset_sigint(void)
-{
-    sigint_received = false;
-}
-
 #if YASH_ENABLE_LINEEDIT
 
 #ifdef SIGWINCH
@@ -1137,7 +1157,7 @@ int trap_builtin(int argc, void **argv)
 		return print_builtin_help(ARGV(0));
 #endif
 	    default:
-		return special_builtin_syntax_error(Exit_ERROR);
+		return special_builtin_error(Exit_ERROR);
 	}
     }
 
@@ -1151,7 +1171,7 @@ int trap_builtin(int argc, void **argv)
 	    if (!sigismember(&printed, s->no)) {
 		sigaddset(&printed, s->no);
 		if (!print_trap(s->name, trap_command[sigindex(s->no)]))
-		    return Exit_FAILURE;
+		    return special_builtin_error(Exit_FAILURE);
 	    }
 	}
 #if defined SIGRTMIN && defined SIGRTMAX
@@ -1160,7 +1180,7 @@ int trap_builtin(int argc, void **argv)
 	    if (sigrtmin + i > sigrtmax)
 		break;
 	    if (!print_trap(get_signal_name(sigrtmin + i), rttrap_command[i]))
-		return Exit_FAILURE;
+		return special_builtin_error(Exit_FAILURE);
 	}
 #endif
     } else if (print) {
@@ -1181,12 +1201,12 @@ int trap_builtin(int argc, void **argv)
 		int index = signum - sigrtmin;
 		if (index < RTSIZE)
 		    if (!print_trap(name, rttrap_command[index]))
-			return Exit_FAILURE;
+			return special_builtin_error(Exit_FAILURE);
 	    } else
 #endif
 	    {
 		if (!print_trap(name, trap_command[sigindex(signum)]))
-		    return Exit_FAILURE;
+		    return special_builtin_error(Exit_FAILURE);
 	    }
 	} while (++xoptind < argc);
     } else {
@@ -1201,8 +1221,7 @@ int trap_builtin(int argc, void **argv)
 	} else {
 	    command = ARGV(xoptind++);
 	    if (xoptind == argc)
-		return special_builtin_syntax_error(
-			insufficient_operands_error(2));
+		return special_builtin_error(insufficient_operands_error(2));
 
 	    if (wcscmp(command, L"-") == 0)
 		command = NULL;
